;;; tab-bar.el --- Tab bar mode for Emacs

;; Copyright (C) 2008  Juri Linkov <juri@jurta.org>

;; Author: Juri Linkov <juri@jurta.org>
;; Keywords: frames
;; Version: 1.0

;; This program is free software; you can redistribute it and/or modify
;; it under the terms of the GNU General Public License as published by
;; the Free Software Foundation, either version 3 of the License, or
;; (at your option) any later version.

;; This program is distributed in the hope that it will be useful,
;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;; GNU General Public License for more details.

;; You should have received a copy of the GNU General Public License
;; along with this program.  If not, see <http://www.gnu.org/licenses/>.

;;; Commentary:

;;

;;; Code:

(defvar tab-bar-image-tab
  "iVBORw0KGgoAAAANSUhEUgAAAEkAAAAbCAIAAAC/VzV3AAAAAXNSR0IArs4c6QAAAJ9JREFUWMPt
2KENgyEUBOBHW88ArIAgeQmBtRgEyx4sAQLBJkgkFX9Kk+oa/ty5J78cZxBjDPqklNJ7p2NjjPHe
7/P1AwshnGuLMRLR5n1tF2zOea7NOVdr3bYH3TewHRIhBHqDDTbsDb3Bhr2hN9hgw97wJmHD3tAb
bH+ypZSY+egPPCLSWjNzSuk6n1LK1hozW2tv0JVSaq2Vc26tvQFQli+WwdTC9AAAAABJRU5ErkJg
gg==")

(defvar tab-bar-image-tab-create
  "iVBORw0KGgoAAAANSUhEUgAAABsAAAAaCAIAAADJ6rCnAAAAAXNSR0IArs4c6QAAAghJREFUSMfV
lj1oE2EYx/+XBHS4NaCXoDTLgfFopR8SweLaUIwnLgYKQWuR21s/Stpi6GBxvsF0C9wkjQUhq9Sh
oDUhOdNBNEVNz48MGUwGpfc+Dgf2WnPJ5cjiM73v/+X98bzP//3ims0mBho+DDr+B2LA3nn9ZkfX
qx4okhSdGB87TrRwd2bnPBCz608BWNBDoq5Xb9+a/dluO00L5YMA9uVGKB8k3wkjURc2wwCMRD1x
7frm843jRABERESOmSgCto1QPghFMM2wkA1jBGz8PH0/MqsfompAmUFsGDgX8AN3d4GyX81Rwivx
89VvZ9RTuJBDbAcAtuMowZAbZJpdicyRePbFaSgzwDDwAwBiTxArC2rw0/RXRyIjYsS61jGHUg6K
cHAg+LNljMCcG2X7jBjzWkegFq9H1LCfa+xN7Q0VhgKltxR3V8f04oJ9KLO69nHqC4D0w3ngJgC8
mgduZFbXQNS7jkvpe5qmtVotSy8UCunFhUeZx046c+k1z/NWQxRFa9RJ77HqleUH9kQARCIRWZaX
l+531LsR/3qdTCad7GkfPaaiKDLq7jUjTdNqtVpH3L86z/O/fvdadaVSKRaL7i+eiYuXXe2eVCrl
kljd/WAn+vrY4Q7R22vOd3J07JL+7n1f0A5ESYrmN55NTl7xcIdvbb2UpKjV5uzv9UDeGW7gP4A/
UXFisiIGSiwAAAAASUVORK5CYII=")

(defvar tab-bar-image-tab-close
  "iVBORw0KGgoAAAANSUhEUgAAABsAAAAaCAIAAADJ6rCnAAAAAXNSR0IArs4c6QAAAhNJREFUSMdj
fP/+PQNVARMDtcFQMJEFmXPq9JnLl6+SYYqurraZqQm6iRDjUlPSyDBx9pxZDAwMEEMRJl6+fDU5
KeXz168MDAzSzN8ZPr19yi0DkeK8eYHx5qVvfnG4TPQPCNq4YR3ERJRw/A8DDDdOM1w8LHnv1P//
/3nfPuNfPlFg/bz/eAH2cITL/ZfTYLx/g2n/OomTB5luXmF8+OirUwiyNjSALIXdjY9ZhH/r2f3/
/Jnl1EGmu7d/OPi/i8ok143/oHL/mNgYXr1heP2O4c0Hxpfv4OKkufHf/////v/79/+f5PObbLM7
GF+8/qui/4+Rm33OfPGmeogUVvT/3z8CvmbZsIDxxpXfzoEPKyf8LGxg/P2ffdliimLmj0soo5J+
/P5TDOdKGRgYZvVPZr56BSJVW12GrKW5tQvN19jDMWHukmXLls2K/AIR3759+/p3r5v+/a+rLV+2
bNmXLwjx2uqypubOfwTdCAE8PDwQhrq6OrJDMMUJ+LqhvhLZIQwMDEpKSoGBgfV1FVjF8ZkIiWsI
++vXr8hSEOdERUVhKSb0TZDjGks4Llu27N69e2jaeHh4WlpasCbGn78I+frSpUvnzp0jvuAxM7f9
T0zMJCQkEGni1Wt3CORrUgtHwnHNyMRhbGJ1+cotkgzFYqKurvb6dWvs7BzIKMMPHTqgq6sNYTMi
19dUqWcYqd4CAADTF+GOC37qSAAAAABJRU5ErkJggg==")

(defvar tab-bar-map-prefix "\C-z")

(defvar tab-bar-prefix-map
  (let ((map (make-sparse-keymap)))
    (define-key map tab-bar-map-prefix (global-key-binding tab-bar-map-prefix))
    map))

(defvar tab-bar-menu-bar-map
  (let ((map (make-sparse-keymap)))
    (define-key map [tabs] (cons "Tabs" tab-bar-prefix-map))
    map))

(defvar tab-bar-map
  (let ((map (make-sparse-keymap)))
    (define-key map tab-bar-map-prefix tab-bar-prefix-map)
    (define-key map [menu-bar] tab-bar-menu-bar-map)
    map)
  "Keymap for the tab bar.")

(define-minor-mode tab-bar-mode
  "Toggle use of the tab bar.
With numeric ARG, display the tab bar if and only if ARG is positive."
  :global t
  :group 'frames
  :keymap tab-bar-map
  (and (display-images-p)
       (tool-bar-mode 1)
       (if (and tab-bar-mode (display-graphic-p))
	   (tab-bar-setup))))

(defvar tab-bar-tabs nil)
(defvar tab-bar-tab-current nil)
(defvar tab-bar-tab-index ?0)

(defvar tab-bar-image-tab-file nil)
(defvar tab-bar-image-tab-create-data nil)
(defvar tab-bar-image-tab-close-data nil)

(defun tab-bar-setup ()
  (let ((coding-system-for-write 'no-conversion))
    (with-temp-file (setq tab-bar-image-tab-file (make-temp-file "tab"))
      (insert (base64-decode-string tab-bar-image-tab)))
    (setq tab-bar-image-tab-create-data (base64-decode-string tab-bar-image-tab-create))
    (setq tab-bar-image-tab-close-data (base64-decode-string tab-bar-image-tab-close)))
  (define-key-after tab-bar-prefix-map (vector ?t)
    `(menu-item "Create tab" tab-bar-tab-create
                :image ,(create-image tab-bar-image-tab-create-data 'png t)))
  (tab-bar-tab-create)
  (define-key-after tab-bar-prefix-map (vector ?c)
    `(menu-item "Close tab"
                (lambda ()
                  (interactive)
                  (define-key tab-bar-prefix-map (vector tab-bar-tab-current) nil)
                  (redraw-frame (selected-frame)))
                :image ,(create-image tab-bar-image-tab-close-data 'png t)))
  (setq tool-bar-map tab-bar-prefix-map))

(defun tab-bar-tab-create ()
  (interactive)
  (setq tab-bar-tab-current (setq tab-bar-tab-index (1+ tab-bar-tab-index)))
  (let ((form '(setq tab-bar-tabs
                     (cons (list tab-bar-tab-current
                                 (buffer-name (current-buffer))
                                 (current-window-configuration)
                                 (delq nil (mapcar
                                            (lambda (b) (and (buffer-live-p b) b))
                                            (frame-parameter (selected-frame)
                                                             'buffer-list)))
                                 (delq nil (mapcar
                                            (lambda (b) (and (buffer-live-p b) b))
                                            (frame-parameter (selected-frame)
                                                             'buried-buffer-list))))
                           (assq-delete-all tab-bar-tab-current tab-bar-tabs)))))
    (eval form)
    (define-key-after tab-bar-prefix-map (vector tab-bar-tab-index)
      `(menu-item ,(buffer-name (current-buffer))
                  ,`(lambda ()
                      (interactive)
                      ,form
                      (let ((current (assq (setq tab-bar-tab-current
                                                 ,tab-bar-tab-index) tab-bar-tabs)))
                        (set-window-configuration (nth 2 current))
                        (modify-frame-parameters
                         (selected-frame) (list (cons 'buffer-list
                                                      (delq nil
                                                            (mapcar (lambda (b)
                                                                      (and (buffer-live-p b) b))
                                                                    (nth 3 current))))))
                        (modify-frame-parameters
                         (selected-frame) (list (cons 'buried-buffer-list
                                                      (delq nil
                                                            (mapcar (lambda (b)
                                                                      (and (buffer-live-p b) b))
                                                                    (nth 4 current)))))))
                      (redraw-frame (selected-frame)))
                  :image ,(tab-bar-create-tab-image (buffer-name (current-buffer)))
                  :enable ,`(not (equal tab-bar-tab-current ,tab-bar-tab-index))
                  :help ,(mapconcat (lambda (w) (buffer-name (window-buffer w)))
                                    (window-list) ", "))
      (1- tab-bar-tab-index)))
  (redraw-frame (selected-frame)))

(defvar tab-bar-image-cache nil)

(defun tab-bar-create-tab-image (name)
  (or (cdr (assoc name tab-bar-image-cache))
      (let* ((coding-system-for-read 'no-conversion)
             (tab-image
              (create-image
               (string-make-unibyte
                (shell-command-to-string
                 (format
                  "convert -background '#f0f0f0f0' -fill black -gravity center -size 73x27 caption:\"%s\" +size %s +swap -gravity south -composite png:-"
                  name tab-bar-image-tab-file)))
               'png t)))
        (push (cons name tab-image) tab-bar-image-cache)
        tab-image)))

(provide 'tab-bar)
;;; tab-bar.el ends here
